import os
import sys
import time
import numpy as np
import matplotlib.pyplot as plt
from uncertainties import ufloat
from scipy.interpolate import interp1d,interp2d
import matplotlib.cm as cm
from scipy.optimize import curve_fit

from .process_raw import *
from .cal_layout import radial_stagger
from .cal_field import *
from .cal_sun import *
from .gen_yaml import gen_yaml, Sun
from .gen_vtk import *
from .input import Parameters
from .output_motab import output_matadata_motab, output_motab
from .master import *


class CRS:
	'''
	The Central Receiver System (CRS) model includes three parts: 
	the sun, the field and the receiver.
	'''

	def __init__(self, latitude, casedir):
		'''
		Arguements:
			casedir : str, the directory of the case 
		'''
		self.casedir=casedir

		if not os.path.exists(casedir):
			os.makedirs(casedir)
		self.latitude=latitude
		self.sun=SunPosition()
		self.master=Master(casedir)

	def receiversystem(self, receiver, rec_w=0., rec_h=0., rec_x=0., rec_y=0., rec_z=100., rec_tilt=0., rec_grid_w=10, rec_grid_h=10, rec_abs=1.):

		'''
		Arguements:
		    (1) receiver  :   str, 'flat', 'cylinder' or directory of the 'stl', type of the receiver
		    (2) rec_w     : float, width of a flat receiver or radius of a cylindrical receiver (m) 
		    (3) rec_h     : float, height of the receiver (m)
		    (4) rec_x     : float, x location of the receiver (m)
		    (5) rec_y     : float, y location of the receiver (m)
		    (6) rec_z     : float, z location of the receiver (m)
		    (7) rec_tilt  : float, tilt angle of the receiver (deg), 0 is where the receiver face to the horizontal
		    (8) rec_grid_w  :   int, number of elements in the horizontal(x)/circumferential direction
		    (9) rec_grid_h  :   int, number of elements in the vertical(z) direction
		    (10) rec_abs   : float, receiver surface absorptivity, e.g. 0.9
		'''
		self.receiver=receiver
		self.rec_abs=rec_abs

		if receiver[-3:]=='stl':
		    self.rec_param=np.r_[rec_w, rec_h, receiver, rec_x, rec_y, rec_z, rec_tilt]
		elif receiver=='flat':
		    self.rec_param=np.r_[rec_w, rec_h, rec_grid_w, rec_grid_h, rec_x, rec_y, rec_z, rec_tilt]
		elif receiver=='cylinder': 
		    self.rec_param=np.r_[rec_w, rec_h, rec_grid_w, rec_grid_h, rec_x, rec_y, rec_z, rec_tilt]  


	def heliostatfield(self, field, hst_rho, slope, hst_w, hst_h, tower_h, tower_r=0.01, hst_z=0., num_hst=0., R1=0., fb=0., dsep=0.):

		'''
		Arguements:
		    (1) field     : str,
		        -- 'polar', 'polar-half', 'surround' or 'surround-half' for desiging a new field 
		        -- the directory of the layout file
		            the layout file is a 'csv' file, (n+2, 7)
		           - n is the total number of heliostats 
		           - the 1st row is the number of each column
		           - the 2nd row is the unit 
		           - the first three columns are X, Y, Z coordinates of each heliostat
		           - the fourth column is the focal length
		           - the last three columns are the aiming points
		    (2) hst_w     : float, width of a heliostat (m) 
		    (3) hst_h     : float, height of a heliostat (m)
		    (4) hst_z     : float, the installation height of the heliostat
		    (5) hst_rho   : float, reflectivity of heliostat 
		    (6) slope     : float, slope error(radians)
		    (7) R1        : float, layout parameter, the distance of the first row of heliostat 
		    (8) dsep      : float, layout parameter, the separation distance of heliostats (m)
		    (9) tower_h    : float, tower height (m)
		    (10)tower_r   : float, radius of tower (m)
		    (11)num_hst  :   int, number of heliostats used in the field design 
		    (12)Q_in_rcv :   int, required heat of the receiver in the field design 
		 '''
	   
		if field[-3:]=='csv':
		    print('KNOWN FIELD')
		    layout=np.loadtxt(field, delimiter=',', skiprows=2)

		else:
		    # design a new field            
		    savefolder=self.casedir+'/des_point'
		    if not os.path.exists(savefolder):
		        os.makedirs(savefolder)

		    pos_and_aiming=radial_stagger(latitude=self.latitude, num_hst=num_hst, width=hst_w, height=hst_h, hst_z=hst_z, towerheight=tower_h, R1=R1, fb=fb, dsep=0., field=field, savedir=savefolder, plot=False)        
		      
		    layout=pos_and_aiming[2:, :]

		self.hst_pos=layout[:,:3].astype(float)
		self.hst_foc=layout[:,3].astype(float) 
		self.hst_aims=layout[:,4:].astype(float)
		self.hst_w=hst_w
		self.hst_h=hst_h
		self.hst_rho=hst_rho
		self.slope=slope
		self.tower_h=tower_h    
		self.tower_r=tower_r

	def yaml(self, dni=1000,sunshape=None,csr=0.01,half_angle_deg=0.2664,std_dev=0.2):
		'''
		Generate YAML files for the Solstice simulation
		'''
		outfile_yaml = self.master.in_case(self.casedir, 'input.yaml')
		outfile_recv = self.master.in_case(self.casedir, 'input-rcv.yaml')

		#att_factor=self.get_attenuation_factor()
		att_factor=1e-6
		print('attenuation', att_factor)
		sun = Sun(sunshape=sunshape, csr=csr, half_angle_deg=half_angle_deg, std_dev=std_dev)

		if self.latitude>0:
			hemisphere='North'
		else:
			hemisphere='South'
		gen_yaml(sun, self.hst_pos, self.hst_foc, self.hst_aims, self.hst_w
		, self.hst_h, self.hst_rho, self.slope, self.receiver, self.rec_param
		, self.rec_abs, outfile_yaml=outfile_yaml, outfile_recv=outfile_recv
		, hemisphere='North', tower_h=self.tower_h, tower_r=self.tower_r
		, spectral=False , medium=att_factor, one_heliostat=False)


	def field_design_annual(self,  dni_des, num_rays, nd, nh, weafile, method, Q_in_des=None, n_helios=None, zipfiles=False, gen_vtk=False, plot=False):
		'''
		Design a field according to the ranked annual performance of heliostats 
		(DNI weighted)

		'''  
		print('')
		print('Start field design')	
		dni_weight=self.dni_TMY(weafile, nd, nh)

		AZI, ZENITH,table,case_list=self.sun.annual_angles(self.latitude, casefolder=self.casedir, nd=nd, nh=nh)
		case_list=case_list[1:]
		SOLSTICE_AZI, SOLSTICE_ELE=self.sun.convert_convention('solstice', AZI, ZENITH)

		oelt=table
		run=np.r_[0]
		nhst=len(self.hst_pos) 
		ANNUAL=N.zeros(nhst)    
		hst_annual={}

		for i in range(len(case_list)):    
			c=int(case_list[i,0].astype(float))
			if c not in run:
				azimuth=SOLSTICE_AZI[c-1]
				elevation= SOLSTICE_ELE[c-1]

				if np.sin(elevation*np.pi/180.)>=1.e-5:
				    dni=1618.*np.exp(-0.606/(np.sin(elevation*np.pi/180.)**0.491))
				else:
				    dni=0.

				sys.stderr.write("\n"+green('Sun position: %s \n'%c))
				print('azimuth: %.2f'% azimuth, ', elevation: %.2f'%elevation)

				onesunfolder=os.path.join(self.casedir,'sunpos_%s'%(c))

				# run solstice
				if elevation<1.:
				    efficiency_total=0
				    performance_hst=np.zeros((nhst, 9))  
				    efficiency_hst=np.zeros(nhst)
				else:
					efficiency_total, performance_hst=self.master.run(azimuth, elevation, num_rays, self.hst_rho, dni, folder=onesunfolder, gen_vtk=False, printresult=False)

					efficiency_hst=performance_hst[:,-1]/performance_hst[:,0]


				hst_annual[c]=performance_hst
				sys.stderr.write(yellow("Total efficiency: {:f}\n".format(efficiency_total)))

			cc=0
			for a in range(len(table[3:])):
			    for b in range(len(table[0,3:])):
			        val=re.findall(r'\d+', table[a+3,b+3])
			        if val==[]:
			            table[a+3,b+3]=0
			        else:
			            if c==float(val[0]):
			                #table[a+3,b+3]=efficiency_total # this line cause wired problem
			                if cc==0: # avoid adding the symetrical point
			                    ANNUAL+=dni_weight[a,b]*efficiency_hst
			                    cc+=1
			run=np.append(run,c)               
		np.savetxt(self.casedir+'/annual_hst.csv',ANNUAL, fmt='%.2f', delimiter=',')
		
		designfolder=self.casedir+'/des_point'
		day=self.sun.days(21, 'Mar')
		dec=self.sun.declination(day)
		hra=0. # solar noon
		zen=self.sun.zenith(self.latitude, dec, hra)
		azi=self.sun.azimuth(self.latitude, zen, dec, hra)        
		azi_des, ele_des=self.sun.convert_convention('solstice', azi, zen) 

		sys.stderr.write("\n"+green('Design Point: \n'))		
		efficiency_total, performance_hst_des=self.master.run(azi_des, ele_des, num_rays, self.hst_rho, dni_des, folder=designfolder, gen_vtk=False, printresult=False)

		Qin=performance_hst_des[:,-1]
		Qsolar=performance_hst_des[0,0]

		ID=ANNUAL.argsort()
		ID=ID[::-1]

		select_hst=np.array([])

		if method==1:
			power=0.
			self.Q_in_rcv=Q_in_des
			for i in range(len(ID)):
			    if power<Q_in_des:
			        idx=ID[i]
			        select_hst=np.append(select_hst, idx)
			        power+=Qin[idx]
			print('')			
			print('Method 1')

		else:
			num_hst=0
			power=0.
			for i in range(len(ID)):
			    if num_hst<n_helios:
			        idx=ID[i]
			        select_hst=np.append(select_hst, idx)
			        num_hst+=1
			        power+=Qin[idx]

			self.Q_in_rcv=power
			print('')			
			print('Method 2')     

		select_hst=select_hst.astype(int)

		self.hst_pos= self.hst_pos[select_hst,:]
		self.hst_foc=self.hst_foc[select_hst]
		self.hst_aims=self.hst_aims[select_hst,:]

		self.n_helios=len(select_hst) 
		self.eff_des=power/float(self.n_helios)/Qsolar

		print('num_hst', self.n_helios)
		print('power   @design', power)
		print('opt_eff @design', self.eff_des)
		Xmax=max(self.hst_pos[:,0])
		Xmin=min(self.hst_pos[:,0])
		Ymax=max(self.hst_pos[:,1])
		Ymin=min(self.hst_pos[:,1])
		A_land=(Xmax-Xmin)*(Ymax-Ymin)
		print('land area', A_land)

		num_hst=len(self.hst_pos)
		title=np.array([['x', 'y', 'z', 'foc', 'aim x', 'aim y', 'aim z'], ['m', 'm', 'm', 'm', 'm', 'm', 'm']])
		design_pos_and_aim=np.hstack((self.hst_pos, self.hst_foc.reshape(num_hst, 1)))
		design_pos_and_aim=np.hstack((design_pos_and_aim, self.hst_aims))
		#symmetric=design_pos_and_aim
		#symmetric[:, 0]=-symmetric[:, 0]
		#design_pos_and_aim=np.vstack((design_pos_and_aim, symmetric))
		#designed_field=design_pos_and_aim
		design_pos_and_aim=np.vstack((title, design_pos_and_aim))
		np.savetxt(self.casedir+'/pos_and_aiming.csv', design_pos_and_aim, fmt='%s', delimiter=',')
		np.savetxt(self.casedir+'/selected_hst.csv', select_hst, fmt='%.0f', delimiter=',')


		# lookup table
		run=np.r_[0]
	
		for i in range(len(case_list)):    
			c=int(case_list[i,0].astype(float))
			if c not in run:                
			    #sundir=designfolder+'/sunpos_%s'%c
			    res_hst=hst_annual[c]
			    Qtot=res_hst[:,0]
			    Qin=res_hst[:,-1]
			    eff=np.sum(Qin[select_hst])/np.sum(Qtot[select_hst])
			    print('')
			    print('sun position:', (c), 'eff', eff)

			for a in range(len(oelt[3:])):
			    for b in range(len(oelt[0,3:])):
			        val=re.findall(r'\d+',oelt[a+3,b+3])
			        if val==[]:
			            oelt[a+3,b+3]=0
			        else:
			            if c==float(val[0]):
			                oelt[a+3,b+3]=eff

		np.savetxt(self.casedir+'/lookup_table.csv', oelt, fmt='%s', delimiter=',')

		return oelt, A_land
		

	def dni_TMY(self, weafile, nd, nh):
		'''
		Argument:

		weafile : str, directory of the weather file .motab
		# col0 - time (s)
		# col1 - GHI
		# col2 -DNI (W/m2)
		# col3 -DHI 
			...
		'''
		with open(weafile) as f:
			content=f.read().splitlines()
		f.close()
	
		lines=len(content)
		seconds=np.array([])
		dni=np.array([])
		for i in range(2, lines-1):
			l=content[i].split(",") 
			seconds=np.append(seconds, l[0])         
			dni=np.append(dni, l[2])
		seconds=seconds.astype(float)
		days=seconds/3600/24
		wea_dec=np.array([])
		for d in days:
			wea_dec=np.append(wea_dec, self.sun.declination(d)) #deg

		wea_hra=((seconds/3600.)%24-12.)*15. #deg
		wea_dni=dni.astype(float)

		hra_lim=180.*(float(nh)/float(nh-1))
		dec_lim=23.45*(float(nd)/float(nd-1))
		hra_bin=np.linspace(-hra_lim, hra_lim, nh+1) 
		dec_bin=np.linspace(-dec_lim, dec_lim, nd+1)
		bins=np.array([hra_bin, dec_bin])


		dni_weight, xbins, ybins=np.histogram2d(wea_hra, wea_dec, bins, weights=wea_dni)

		return dni_weight.T

	def get_attenuation_factor(self):

		foc=self.hst_foc.astype(float)

		# to get the attenuation factor
		def func(x, b):
		    return np.exp(-b * x)
		def fun_two(x):
		    return 0.99321-0.0001176*x+1.97e-8*x**2
		xdata = np.linspace(0, np.max(foc), np.max(foc)*100)
		y = fun_two(xdata)
		ydata = y
		popt, pcov = curve_fit(func, xdata, ydata)
		y2 = [func(i, popt[0]) for i in xdata]
		att_factor =popt[0]
		return att_factor

	def generateVTK(self,eta_hst, savevtk):

		FPF=FieldPF(0., 0., np.r_[0., 1., 0.])
		norms=np.zeros(np.shape(self.hst_pos))
		norms[:,-1]=1.
		COORD, TRI, element, nc=FPF.view_heliostats(width=self.hst_w, height=self.hst_h, normals=norms, hstpos=self.hst_pos)
		NORMS=np.repeat(norms, element, axis=0)

		#field performance
		hst_tot  =eta_hst[:,0]
		rec_abs=eta_hst[:,1]

		hst_eff=rec_abs/hst_tot 

		savevtk=savevtk+'/results-field.vtk'
		TOT=np.repeat(hst_tot, element) 
		ABS=np.repeat(rec_abs, element)   
		EFF=np.repeat(hst_eff, element) 

		DATA={'tot': TOT, 'rec_abs': ABS, 'efficiency':EFF}
		gen_vtk(savedir=savevtk, points=COORD.T, indices=TRI, norms=NORMS, colormap=True, DATA=DATA)

if __name__=='__main__':
	start=time.time()
	casedir='./test-crs-design'
	tablefile=casedir+'/OELT_Solstice.motab'
	if os.path.exists(tablefile):    
		print('')
		print('Load exsiting OELT')

	else:

		pm=Parameters()
		pm.Q_in_rcv=565e6
		pm.nd=5
		pm.nh=5
		pm.H_tower=250.
		pm.H_rcv=24.
		pm.W_rcv=24.
		pm.dependent_par()
		pm.saveparam(casedir)
		print(pm.fb)
		print(pm.H_tower)
		crs=CRS(latitude=pm.lat, casedir=casedir)   
		weafile='/home/yewang/solartherm-master/SolarTherm/Data/Weather/gen3p3_Daggett_TMY3.motab'
		crs.heliostatfield(field=pm.field_type, hst_rho=pm.rho_helio, slope=pm.slope_error, hst_w=pm.W_helio, hst_h=pm.H_helio, tower_h=pm.H_tower, tower_r=pm.R_tower, hst_z=pm.Z_helio, num_hst=pm.n_helios, R1=pm.R1, fb=pm.fb, dsep=pm.dsep)

		crs.receiversystem(receiver=pm.rcv_type, rec_w=float(pm.W_rcv), rec_h=float(pm.H_rcv), rec_x=float(pm.X_rcv), rec_y=float(pm.Y_rcv), rec_z=float(pm.Z_rcv), rec_tilt=float(pm.tilt_rcv), rec_grid_w=int(pm.n_W_rcv), rec_grid_h=int(pm.n_H_rcv),rec_abs=float(pm.alpha_rcv))

		crs.yaml(sunshape=pm.sunshape,csr=pm.crs,half_angle_deg=pm.half_angle_deg,std_dev=pm.std_dev)

		oelt, A_land=crs.field_design_annual(dni_des=900., num_rays=int(1e6), nd=pm.nd, nh=pm.nh, weafile=weafile, method=1, Q_in_des=pm.Q_in_rcv, n_helios=None, zipfiles=False, gen_vtk=False, plot=False)


		if (A_land==0):    
		    tablefile=None
		else:                                                
		    A_helio=pm.H_helio*pm.W_helio
		    output_matadata_motab(table=oelt, field_type=pm.field_type, aiming='single', n_helios=self.n_helios, A_helio=A_helio, eff_design=self.eff_des, H_rcv=pm.H_rcv, W_rcv=pm.W_rcv, H_tower=pm.H_tower, Q_in_rcv=pm.Q_in_rcv, A_land=A_land, savedir=tablefile)


	end=time.time()
	print('total time %.2f'%((end-start)/60.), 'min')






import numpy as N
import re
import sys
import os
from uncertainties import ufloat
from uncertainties.umath import * 

def process_raw_results(rawfile, savedir,rho_mirror,dni):
	"""Process the raw Solstice `simul` output into readable CSV files

	``Arguments``

	  * rawfile (str): the directory of the `simul` file that generated by Solstice
	  * savedir (str): the directory for saving the organised results
	  * rho_mirror (float): mirror reflectivity (needed for reporting energy sums)
	  * dni (float): the direct normal irradiance (W/m2), required to obtain performance of individual heliostat

	``Returns``

	  * efficiency_total (float): the total optical efficiency
	  * performance_hst (numpy array): the breakdown of losses of each individual heliostat
	  * The simulation results are created and written in the `savedir`
		
	"""

	# FIXME this approach seems fundamentally a bit messy... we are carefully
	# load the text output from 'solppraw' and then assuming that we can
	# correctly find the right bits of data in the right place. Wouldn't it
	# be easier and more robust to modify solppraw to create output that can
	# be directly loaded, along with data labels, eg a YAML file? Or to
	# create 'result-raw.csv' directly?

	rows = []
	index = 0
	with open(rawfile) as f:
		for r in f.readlines():
			if index<20:
				pass #sys.stderr.write("Line %d: %s"%(index,r))

			if r[0] == "#":
				#sys.stderr.write("^------Comment\n")
				#comment line
				rows.append([r])
			else:
				rows.append(r.split())			
			index+=1

	results=N.array([])

	# sun direction

	#sys.stderr.write("SUN DIRECTION?\n")
	#sys.stderr.write(rows[0][0]+"\n")
	sun=re.findall("[-+]?\d*\.\d+|\d+", rows[0][0])
	#sys.stderr.write("SUN = %s\n"%(repr(sun),))

	azimuth=sun[0]
	elevation=sun[1]

	def get_rc(row,col):
		return float(rows[row][col])

	# Global results

	num_res=int(get_rc(1,0)) # number of global results
	num_rec=int(get_rc(1,1))
	num_hst=get_rc(1,2)
	num_rays=get_rc(1,3)

	potential=get_rc(2,0) #W
	potential_err=get_rc(2,1)

	absorbed=get_rc(3,0)
	absorbed_err=get_rc(3,1)    

	Fcos=get_rc(4,0)
	Fcos_err=get_rc(4,1)  

	shadow_loss=get_rc(5,0)
	shadow_err=get_rc(5,1)    

	missing_loss=get_rc(6,0)
	missing_err=get_rc(6,1) 

	material_loss=get_rc(7,0)
	material_err=get_rc(7,1)    

	atmospheric_loss=get_rc(8,0)
	atmospheric_err=get_rc(8,1)  

	# Target (receiver)
	# 0 receiver name
	# 1 - 2 id and area
	# 3 - 24 (total 22) front
	# 25- 46 (total 22) back 
	rec_area=get_rc(num_res+2,2) # m2  
	
	rec_front_income=get_rc(num_res+2,3)
	rec_front_income_err=get_rc(num_res+2,4)
	#rec_no_material_loss=get_rc(num_res+2,5)
	#rec_no_material_loss_err=get_rc(num_res+2,6)
	#rec_no_atmo_loss=get_rc(num_res+2,7)
	#rec_no_atmo_loss_err=get_rc(num_res+2,8)
	#rec_material_loss=get_rc(num_res+2,9)
	#rec_material_loss_err=get_rc(num_res+2,10)
	rec_front_absorbed=get_rc(num_res+2,13)
	rec_front_absorbed_err=get_rc(num_res+2,14)
	rec_front_eff=get_rc(num_res+2,23)
	rec_front_eff_err=get_rc(num_res+2,24)

	rec_back_income=get_rc(num_res+2,25)
	rec_back_income_err=get_rc(num_res+2,26)
	rec_back_absorbed=get_rc(num_res+2,35)
	rec_back_absorbed_err=get_rc(num_res+2,36)
	rec_back_eff=get_rc(num_res+2,-2)
	rec_back_eff_err=get_rc(num_res+2,-1)

	if num_rec>1:
		#Virtual target
		vir_area=get_rc(num_res+3,2)
		vir_income=get_rc(num_res+3,3)
		vir_income_err=get_rc(num_res+3,4)
	
	raw_res=N.array([
		['name','value', 'error']
		,['sun_azimuth', azimuth,'']
		,['sun_elevation', elevation, '']
		,['num hst', num_hst,'']
		,['num rays',num_rays, '']
		,['potential flux', potential, potential_err]
		,['absorbed flux', absorbed, absorbed_err]
		,['Cosine factor', Fcos, Fcos_err]
		,['shadow loss', shadow_loss, shadow_err]
		,['Mising loss', missing_loss, missing_err]
		,['materials loss', material_loss, material_err]
		,['atomospheric loss', atmospheric_loss, atmospheric_err]
		,['','','']
		,['Target', '','']
		,['area', rec_area, '']
		,['front income flux', rec_front_income, rec_front_income_err]
		,['back income flux', rec_back_income, rec_back_income_err]
		,['front absorbed flux', rec_front_absorbed, rec_front_absorbed_err]
		,['back absorbed flux', rec_back_absorbed, rec_back_absorbed_err]
		,['front efficiency', rec_front_eff, rec_front_eff_err]
		,['back efficiency', rec_back_eff, rec_back_eff_err]
		,['','','']
		,['Virtual plane','','']
		,['area', vir_area, '']
		,['income flux', vir_income,vir_income_err]
	])

	#sys.stderr.write(repr(raw_res))
	#sys.stderr.write("SHAPE = %s" % (repr(raw_res.shape)))

	N.savetxt(savedir+'/result-raw.csv', raw_res, fmt='%s', delimiter=',')

	Qtotal=ufloat(potential, 0)
	Fcos=ufloat(Fcos,Fcos_err) 
	Qcos=Qtotal*(1.-Fcos)
	Qshade=ufloat(shadow_loss,shadow_err)
	Qfield_abs=(Qtotal-Qcos-Qshade)*(1.-float(rho_mirror))
	Qatm=ufloat(atmospheric_loss, atmospheric_err)
	Qspil=ufloat(vir_income,vir_income_err)
	Qabs=ufloat(absorbed, absorbed_err)
	Qrefl=ufloat(rec_front_income,rec_front_income_err)+ufloat(rec_back_income,rec_back_income_err)-Qabs
	Qblock=Qtotal-Qcos-Qshade-Qfield_abs-Qspil-Qabs-Qrefl-Qatm

	organised=N.array([
		['Name', 'Value', '+/-Error']
		,['Qall (kW)', Qtotal.n/1000., Qtotal.s/1000.]
		,['Qcos (kW)', Qcos.n/1000.,Qcos.s/1000.]
		,['Qshad (kW)', Qshade.n/1000., Qshade.s/1000.]
		,['Qfield_abs (kW)', Qfield_abs.n/1000., Qfield_abs.s/1000.]
		,['Qblcok (kW)', Qblock.n/1000.,  Qblock.s/1000.]
		,['Qatm (kW)',Qatm.n/1000., Qatm.s/1000.]
		,['Qspil (kW)', Qspil.n/1000., Qspil.s/1000.]
		,['Qrefl (kW)', Qrefl.n/1000.,Qrefl.s/1000.]
		,['Qabs (kW)', Qabs.n/1000., Qabs.s/1000.]
		,['rays', num_rays,'-']
	])
	N.savetxt(savedir+'/result-formatted.csv', organised, fmt='%s', delimiter=',')
	efficiency_total=Qabs/Qtotal

	# per heliostat results, and
	# per receiver per heliostat results
	num_hst=int(num_hst)    
	heliostats=N.zeros((num_hst,28))

	for i in range(num_hst):
		l1=2+num_res+num_rec+i # the line number of the per heliostat result
		per_hst=rows[l1]

		hst_idx=re.findall("[-+]?\d*\.\d+|\d+", per_hst[0] ) 
		hst_area=per_hst[2]
		hst_sample=per_hst[3]
		hst_cos=per_hst[4]
		hst_shad=per_hst[6]

		heliostats[i,0]=hst_idx[0]
		heliostats[i,1]=hst_area
		heliostats[i,2]=hst_sample
		heliostats[i,3]=hst_cos
		heliostats[i,4]=hst_shad

		# per heliostat per receiver
		l2=2+num_res+num_rec+num_hst+i  
		per_hst=rows[l2]   
		hst_in=float(per_hst[2])+float(per_hst[22]) # front+back
		hst_in_mat=float(per_hst[8])+float(per_hst[28])
		hst_in_atm=float(per_hst[10])+float(per_hst[30])
		hst_abs=float(per_hst[12])+float(per_hst[32])
		hst_abs_mat=float(per_hst[18])+float(per_hst[38])
		hst_abs_atm=float(per_hst[20])+float(per_hst[40])
		  
		heliostats[i,5]=hst_in
		heliostats[i,6]=hst_in_mat
		heliostats[i,7]=hst_in_atm
		heliostats[i,8]=hst_abs
		heliostats[i,9]=hst_abs_mat
		heliostats[i,10]=hst_abs_atm

		# per heliostat per virtual target
		l3=2+num_res+num_rec+2*num_hst+i  
		per_hst=rows[l3] 
		hst_in=float(per_hst[2])+float(per_hst[22]) # front+back
		hst_in_mat=float(per_hst[8])+float(per_hst[28])
		hst_in_atm=float(per_hst[10])+float(per_hst[30])
		hst_abs=float(per_hst[12])+float(per_hst[32])
		hst_abs_mat=float(per_hst[18])+float(per_hst[38])
		hst_abs_atm=float(per_hst[20])+float(per_hst[40])
		  
		heliostats[i,11]=hst_in
		heliostats[i,12]=hst_in_mat
		heliostats[i,13]=hst_in_atm
		heliostats[i,14]=hst_abs
		heliostats[i,15]=hst_abs_mat
		heliostats[i,16]=hst_abs_atm


		hst_tot=float(hst_area)*1000.
		hst_cos=hst_tot*(1.-float(hst_cos))
		hst_shad=float(hst_shad)
		hst_abs=(hst_tot-hst_cos-hst_shad)*(1.-rho_mirror)

		hst_atm=float(heliostats[i,16])
		hst_spil=float(heliostats[i,11])
		hst_rec_abs=float(heliostats[i,8])
		hst_rec_refl=float(heliostats[i,5])-float(heliostats[i,8])
		hst_block=hst_tot-hst_cos-hst_shad-hst_abs-hst_atm-hst_spil-hst_rec_abs-hst_rec_refl

		heliostats[i,19]=hst_tot
		heliostats[i,20]=hst_cos
		heliostats[i,21]=hst_shad
		heliostats[i,22]=hst_abs
		heliostats[i,23]=hst_block
		heliostats[i,24]=hst_atm
		heliostats[i,25]=hst_spil
		heliostats[i,26]=hst_rec_refl
		heliostats[i,27]=hst_rec_abs

	idx=heliostats[:, 0].argsort()
	heliostats=heliostats[idx]
	performance_hst=heliostats[:, 19:]

	heliostats_title=N.array(['hst_idx', 'area', 'sample', 'cos', 'shade', 'incoming', 'in-mat-loss','in-atm-loss', 'absorbed', 'abs-mat-loss', 'abs-atm-loss', 'vir_incoming', 'vir_in-mat-loss','vir_in-atm-loss', 'vir_absorbed', 'vir_abs-mat-loss', 'vir_abs-atm-loss', '', '', 'total', 'cos', 'shad', 'hst_abs', 'block', 'atm', 'spil', 'rec_refl', 'rec_abs' ]) 

	heliostats_details=N.vstack((heliostats_title, heliostats))

	N.savetxt(savedir+'/heliostats-raw.csv', heliostats_details, fmt='%s', delimiter=',')


	return efficiency_total, performance_hst

if __name__=='__main__':
    eta,pf_hst = proces_raw_results(sys.argv[1], sys.argv[2], sys.argv[3])
    sys.stderr.write('\nTotal efficiency: %s\n'%(repr(eta),))





import numpy as np
from datetime import datetime

def output_motab(table,savedir=None):
	'''
	output the .motab table fiel
	'''
	f=open(savedir, 'w')
	f.write('#1\n')


	# size of the lookup table
	m=np.shape(table)[0]-2
	n=np.shape(table)[1]-2
	f.write('double optics(%s, %s)\n'%(m,n))

	hour_angle=table[2, 3:]
	declination=table[3:,2]

	for i in range(m):
		if i ==0:
			row_i=np.append(0, hour_angle)
		else:
			row_i=np.append(declination[i-1], table[2+i, 3:])

		#content=np.array2string(row_i, formatter={'float_kind':lambda x: "%.2f" % row_i})
		#content=np.array2string(row_i, precision=2, separator=' ', suppress_small=True)
		#f.write(content+'\n')
		f.write(" ".join(map(str, row_i)))
		f.write("\n")

	f.close()


def output_matadata_motab(table, field_type, aiming, n_helios, A_helio, eff_design, H_rcv, W_rcv, H_tower, Q_in_rcv, A_land, savedir=None):
	'''
	output the .motab table fiel
	'''
	f=open(savedir, 'w')
	f.write('#1\n')
	f.write('#Comments: Field type: %s, Aiming Strategy: %s, Date:%s\n'%(field_type, aiming, datetime.now()))
	f.write('#METALABELS,n_helios,A_helio,Eff_design,H_rcv,W_rcv,H_tower, Q_in_rcv, A_land\n')
	f.write('##METAUNITS,real,m2,real,m,m,m,W,m2\n')
	f.write('#METADATA,%s,%s,%s,%s,%s,%s,%s,%s\n'%(n_helios,A_helio,eff_design,H_rcv,W_rcv,H_tower,Q_in_rcv,A_land))

	# size of the lookup table
	m=np.shape(table)[0]-2
	n=np.shape(table)[1]-2
	f.write('double optics(%s, %s)\n'%(m,n))

	hour_angle=table[2, 3:]
	declination=table[3:,2]

	for i in range(m):
		if i ==0:
			row_i=np.append(0, hour_angle)
		else:
			row_i=np.append(declination[i-1], table[2+i, 3:])

		#content=np.array2string(row_i, formatter={'float_kind':lambda x: "%.2f" % row_i})
		#content=np.array2string(row_i, precision=2, separator=' ', suppress_small=True)
		#f.write(content+'\n')
		f.write(" ".join(map(str, row_i)))
		f.write("\n")

	f.close()


def read_motab(filename):

	with open(filename) as f:
		content=f.read().splitlines()
	f.close()
	res=content[4].split(',')

	n_helios=float(res[1])
	A_helio=float(res[2])
	eff_des=float(res[3])
	Q_in_rcv=float(res[-2])
	A_land=float(res[-1])

	oelt=np.array([])
	solar_hour=np.array([])
	declination=np.array([])
	t=content[6].split(' ')

	for v in t[1:]:
		solar_hour=np.append(solar_hour, float(v))

	for t in content[7:]:
		v=t.split(' ')
		declination=np.append(declination, float(v[0]))
		oelt=np.append(oelt, np.array(v[1:],dtype=float))

	oelt=oelt.reshape(len(declination), len(solar_hour))

	return n_helios, A_helio, eff_des, Q_in_rcv, A_land, solar_hour, declination, oelt
	





# This script finds the location of Solstice from the Windows Registry, assuming
# it was installed with ANU's NSIS-based installer. It works with Python 2.7 and
# Python 3.8, with both 32 and 64-bit versions, on both Windows 7 and Windows 10
# -- John Pye, April 2020

import sys, platform, os

def find_solstice_root(version_required=None,verbose=0):
	"""Locate the place where Solstice files are installed on Windows

	``Arguments``

	  * version_required (None or str): if not None, then enforce this specified version of Solstice (eg '0.9.0')
	  * verbose (bool): if True, output file locations to stderr

	``Return``

	  * dirn (str): the directory where Solstice is installed in the Windows system
	"""

	if verbose: sys.stderr.write("Python is running from %s\n"%(sys.executable,))
	if not platform.system()=="Windows":
		raise RuntimeError("This function is only for Windows")
	# otherwise...
	if sys.version_info[0] < 3:
		if verbose: sys.stderr.write("Python 2, ")
		import _winreg as winreg
	else:
		if verbose: sys.stderr.write("Python 3, ")
		import winreg
	if sys.maxsize > 2**32:
		if verbose: sys.stderr.write("64-bit\n")
		# 64-bit Python
		key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,r"Software\Meso-Star\Solstice",0,winreg.KEY_READ)
	else:
		if verbose: sys.stderr.write("32-bit\n")
		key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,r"Software\Meso-Star\Solstice",0,winreg.KEY_WOW64_64KEY+winreg.KEY_READ)
	dirn, typd = winreg.QueryValueEx(key, 'root')
	ver, typv = winreg.QueryValueEx(key, 'version')
	if version_required:
		assert ver == version_required
	return dirn

def find_prog(name,version_required=None):
	"""Find the path to any required Solstice executable program

	``Arguments``

	  * name (str): stem-name (eg 'solpp') of the program (eg 'solpp.exe') required
	  * version_required (None or str): if not None, then enforce this specified version of Solstice (e.g. '0.9.0')

	``Return``

	  * path (str): path of the required Solstice executable program
	"""
	
	if platform.system()=="Windows":
		path = os.path.join(find_solstice_root(version_required),"bin","%s.exe"%(name,));
		if not os.path.exists(path):
			raise RuntimeError("Program '%s' was not found at path '%s'"%(name,path))
		return path
	else:
		# assume all solstice programs are on the PATH
		import subprocess
		DEVNULL = open(os.devnull, 'wb')
		rc = subprocess.call(['which',name],stdout=DEVNULL)
		DEVNULL.close()
		if rc:
			raise RuntimeError("Program '%s' was not found in the PATH" %(name))
		return name

if __name__=="__main__":
	dirn = find_solstice_root('0.9.0',verbose=1)
	sys.stderr.write("Solstice is installed in %s\n\n" %(dirn,)) # works both python2+3 :o)

	import subprocess, os

	spath = os.path.join(dirn,"bin","solstice.exe")

	# check that installed version of solstice is 0.9.0:
	ret = subprocess.check_output([spath,"--version"])
	assert ret.decode('ascii').strip() == "Solstice 0.9.0"

	# output the help text from solstice:
	subprocess.check_call([spath,"-h"])

# you should see the 'solstice -help' output shown in your console.

# checked with Python 2.7 32-bit on Windows 7 
# checked with Python 3.8.2 32-bit on Windows 10
# checked with Python 3.8.2 64-bit on Windows 10




from __future__ import absolute_import, unicode_literals
from celery import Celery

BROKER = 'amqp://myuser:mypassword@localhost:5672/myvhost'
BACKEND = 'redis://localhost:6379/0'

app = Celery('celery1', backend=BACKEND, broker=BROKER,include=['celsol.tasks'])
app.conf.update(
    broker_transport_options = {'visibility_timeout': 3600} # 1 h
)

if __name__ == '__main__':
    app.start()




import numpy as np
import platform
import os, sys, subprocess, glob, datetime
import colorama
colorama.init()

from .process_raw import *
from .find_solstice import *
from .cal_sun import *

def yellow(text):
    return colorama.Fore.YELLOW + colorama.Style.BRIGHT + text + colorama.Style.RESET_ALL

def green(text):
    return colorama.Fore.GREEN + colorama.Style.BRIGHT + text + colorama.Style.RESET_ALL

def SPROG(name):
    return find_prog(name)

def run_prog(name,args,output_file=None,verbose=True):
	prog = SPROG(name)
	args1 = [str(a) for a in args]
	if verbose: 
		sys.stderr.write("Running '%s' with args: %s\n" % (name," ".join(args1)))
	if output_file is not None:
		# any error will cause an exception (and we capture the output to a file)
		res = subprocess.check_output([prog]+args1)
		with open(output_file,'w') as f:
			f.write(res.decode('ascii'))
	else:
		# any error will cause an exception...
		subprocess.check_call([prog]+args1)

class Master:

	def __init__(self, casedir='.'):
		"""Set up the Solstice simulation, i.e. establishing the case folder, calling the Solstice program and post-processing the results

		``Argument``
		  * casedir (str): the case directory
		"""
		self.casedir=casedir
		if not os.path.exists(self.casedir):
		    os.makedirs(self.casedir)
		    assert os.path.isdir(casedir)
		sys.stderr.write("Case directory is '%s'\n" % (yellow(self.casedir),))


	def in_case(self, folder, fn):
		"""Joining a file name with the case directory

		``Argument``

		  * fn (str): file name

		``Return``

		  * a joining directory of the file in the case directory
		"""
		
		if not os.path.exists(folder):
		    os.makedirs(folder)
		    assert os.path.isdir(folder)

		return os.path.join(folder,fn)

	def run(self, azimuth, elevation, num_rays, rho_mirror, dni, folder, gen_vtk=True, printresult=True):

		"""Run an optical simulation (one sun position) using Solstice 

		* `azimuth` (float): the azimuth angle of the ray-tracing simulation in Solstice, counted from East towards to North
		* `elevation` (float): the elevation angle of the ray-tracing simulation in Solstice
		* `num_rays` (int): number of rays to be cast in the ray-tracing simulation
		* `rho_mirror` (float): reflectivity of mirrors, required for results post-processing 
		* `dni` (float): the direct normal irradiance (W/m2), required to obtain performance of individual heliostat
		* `gen_vtk` (boolean): if True, generate .vtk files for rendering in Paraview
			
		Returns: no return value (results files are created and written)
		"""

		YAML_IN = self.in_case(self.casedir, 'input.yaml')
		RECV_IN = self.in_case(self.casedir, 'input-rcv.yaml')

		# main raytrace
		run_prog("solstice",['-D%s,%s'%(azimuth,elevation),'-v','-n',num_rays,'-R',RECV_IN,'-fo',self.in_case(folder, 'simul'),YAML_IN])


		if gen_vtk:
			dirn = os.getcwd()
			try:
				os.chdir(folder)

				# Read "simul" results and produce a text file with the raw results
				run_prog('solppraw',[self.in_case(folder, 'simul')])

				# post processing
				run_prog("solstice",['-D%s,%s'%(azimuth,elevation),'-g','format=obj:split=geometry','-fo',self.in_case(folder, 'geom'),YAML_IN])

				# run a short raytrace to produce some ray paths
				run_prog("solstice",['-D%s,%s'%(azimuth,elevation),'-q','-n','100','-R',RECV_IN,'-p','default',YAML_IN], output_file=self.in_case(folder, 'solpaths'))

				# Read "simul" results and produce receiver files (.vtk) of incoming and/or absorbed solar flux per-primitive
				run_prog('solmaps',[self.in_case(folder, 'simul')])

				# Read "geom" and "simul" file results and produce primaries and receivers files (.vtk), and .obj geometry files
				run_prog('solpp',[self.in_case(folder, 'geom'),self.in_case(folder, 'simul')])

				# Read "solpaths" file and produce readable file (.vtk) by paraview to visualize the ray paths
				run_prog('solpaths',[self.in_case(folder, 'solpaths')])
			finally:
				os.chdir(dirn)

		eta, performance_hst=process_raw_results(self.in_case(folder, 'simul'), folder,rho_mirror,dni)

		if printresult:
			sys.stderr.write('\n' + yellow("Total efficiency: {:f}\n".format(eta)))
			sys.stderr.write(green("Completed successfully.\n"))

		return eta, performance_hst

	def run_annual(self, nd, nh, latitude, num_rays, num_hst,rho_mirror,dni,gen_vtk=False):

		"""Run a list of optical simulations to obtain annual performance (lookup table) using Solstice 

		``Arguments``

		  * nd (int): number of rows in the lookup table (discretisation of the declination angle)
		  * nh (int): number of columns in the lookup table (discretisation of the solar hour angle)
		  * latitude (float): the latitude angle of the plan location (deg)
		  * num_rays (int): number of rays to be cast in the ray-tracing simulation
		  * num_hst (int): number of heliostats 
		  * rho_mirror (float): reflectivity of mirrors, required for results post-processing 
		  * dni (float): the direct normal irradiance (W/m2), required to obtain performance of individual heliostat
		  * gen_vtk (bool): True - perform postprocessing for visualisation of  each individual ray-tracing scene (each sun position), False - no postprocessing for visualisation 


		``Return``

		  * No return value (results files are created and written)
		"""

		YAML_IN = self.in_case(self.casedir, 'input.yaml')
		RECV_IN = self.in_case(self.casedir, 'input-rcv.yaml')

		sun=SunPosition()
		AZI, ZENITH,table,case_list=sun.annual_angles(latitude, casefolder=self.casedir, nd=nd, nh=nh)
		case_list=case_list[1:]
		SOLSTICE_AZI, SOLSTICE_ELE=sun.convert_convention('solstice', AZI, ZENITH)

		# performance of individual heliostat is recorded
		# TODO note, DNI is not varied in the simulation, 
		# i.e. performance is not dni-weighted
		ANNUAL=np.zeros((num_hst, 9))    
		run=np.r_[0]

		for i in range(len(case_list)):     
			c=int(case_list[i,0].astype(float))
			if c not in run:
				azimuth=SOLSTICE_AZI[c-1]
				elevation=SOLSTICE_ELE[c-1]
				#if np.sin(elevation*np.pi/180.)>=1.e-5:
				#	dni=1618.*np.exp(-0.606/(np.sin(elevation*np.pi/180.)**0.491))
				#else:
				#	dni=0.

			   
				sys.stderr.write("\n"+green('Sun position: %s \n'%c))
				print('azimuth: %.2f'% azimuth, ', elevation: %.2f'%elevation)

				onesunfolder=os.path.join(self.casedir,'sunpos_%s'%(c))

				# run solstice
				if elevation<1.: # 1 degree
				    efficiency_total=ufloat(0,0)
				    performance_hst=np.zeros((num_hst, 9))  
				else:
					efficiency_total, performance_hst=self.run(azimuth, elevation, num_rays, rho_mirror, dni, folder=onesunfolder, gen_vtk=False, printresult=False)

					sys.stderr.write(yellow("Total efficiency: {:f}\n".format(efficiency_total)))

				ANNUAL+=performance_hst
			else:
				ANNUAL+=performance_hst


			for a in range(len(table[3:])):
				for b in range(len(table[0,3:])):
				    val=re.findall(r'\d+',    table[a+3,b+3])
				    if val==[]:
				        table[a+3,b+3]=0
				    else:
				        if c==float(val[0]):
				            table[a+3,b+3]=efficiency_total.nominal_value

			run=np.append(run,c)   

		annual_title=np.array(['Q_solar','Q_cosine', 'Q_shade', 'Q_hst_abs', 'Q_block', 'Q_atm', 'Q_spil', 'Q_refl', 'Q_rcv_abs']) 
		ANNUAL=np.vstack((annual_title, ANNUAL))
		np.savetxt(self.casedir+'/lookup_table.csv', table, fmt='%s', delimiter=',')
		np.savetxt(self.casedir+'/result-heliostats-annual-performance.csv', ANNUAL, fmt='%s', delimiter=',')
		sys.stderr.write("\n"+green("Lookup table saved.\n"))
		sys.stderr.write(green("Completed successfully.\n"+"\n"))
		return table, ANNUAL


